@@ -40,11 +40,12 @@
 #include <bluetooth/sdp.h>
 #include <bluetooth/sdp_lib.h>
 
-#include "logging.h"
+#include "log.h"
 #include "pbap.h"
 #include "sync.h"
 #include "transfer.h"
 #include "session.h"
+#include "btio.h"
 
 #define AGENT_INTERFACE  "org.openobex.Agent"
 
@@ -71,7 +72,7 @@ struct session_callback {
 	void *data;
 };
 
-struct agent_pending {
+struct pending_data {
 	DBusPendingCall *call;
 	session_callback_t cb;
 	struct transfer_data *transfer;
@@ -81,7 +82,7 @@ struct agent_data {
 	char *name;
 	char *path;
 	guint watch;
-	struct agent_pending *pending;
+	struct pending_data *pending;
 };
 
 static void session_prepare_put(struct session_data *session, void *data);
@@ -90,12 +91,12 @@ struct session_data *session_ref(struct session_data *session)
 {
 	g_atomic_int_inc(&session->refcount);
 
-	debug("session_ref(%p): ref=%d", session, session->refcount);
+	DBG("%p: ref=%d", session, session->refcount);
 
 	return session;
 }
 
-static void free_pending(struct agent_pending *pending)
+static void free_pending(struct pending_data *pending)
 {
 	if (pending->call)
 		dbus_pending_call_unref(pending->call);
@@ -156,12 +157,12 @@ static void session_unregistered(struct session_data *session)
 	g_dbus_unregister_interface(session->conn, session->path,
 					SESSION_INTERFACE);
 
-	debug("Session(%p) unregistered %s", session, session->path);
+	DBG("Session(%p) unregistered %s", session, session->path);
 }
 
 static void session_free(struct session_data *session)
 {
-	debug("session_free(%p)", session);
+	DBG("%p", session);
 
 	if (session->agent)
 		agent_release(session);
@@ -194,7 +195,7 @@ void session_unref(struct session_data *session)
 
 	ret = g_atomic_int_dec_and_test(&session->refcount);
 
-	debug("session_unref(%p): ref=%d", session, session->refcount);
+	DBG("%p: ref=%d", session, session->refcount);
 
 	if (ret == FALSE)
 		return;
@@ -202,16 +203,17 @@ void session_unref(struct session_data *session)
 	session_free(session);
 }
 
-static gboolean rfcomm_callback(GIOChannel *io, GIOCondition cond,
-							gpointer user_data)
+static void rfcomm_callback(GIOChannel *io, GError *err, gpointer user_data)
 {
 	struct callback_data *callback = user_data;
 	struct session_data *session = callback->session;
 	GwObex *obex;
 	int fd;
 
-	if (cond & (G_IO_NVAL | G_IO_ERR))
+	if (err != NULL) {
+		error("%s", err->message);
 		goto done;
+	}
 
 	fd = g_io_channel_unix_get_fd(io);
 
@@ -227,63 +229,27 @@ done:
 	session_unref(callback->session);
 
 	g_free(callback);
-
-	return FALSE;
 }
 
 static int rfcomm_connect(const bdaddr_t *src,
 				const bdaddr_t *dst, uint8_t channel,
-					GIOFunc function, gpointer user_data)
+					BtIOConnect function, gpointer user_data)
 {
 	GIOChannel *io;
-	struct sockaddr_rc addr;
-	int sk;
+	GError *err = NULL;
 
-	sk = socket(PF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM);
-	if (sk < 0)
-		return -EIO;
+	io = bt_io_connect(BT_IO_RFCOMM, function, user_data, NULL, &err,
+				BT_IO_OPT_SOURCE_BDADDR, src,
+				BT_IO_OPT_DEST_BDADDR, dst,
+				BT_IO_OPT_CHANNEL, channel,
+				BT_IO_OPT_SEC_LEVEL, BT_IO_SEC_LOW,
+				BT_IO_OPT_INVALID);
+	if (io != NULL)
+		return 0;
 
-	memset(&addr, 0, sizeof(addr));
-	addr.rc_family = AF_BLUETOOTH;
-	bacpy(&addr.rc_bdaddr, src);
-
-	if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
-		close(sk);
-		return -EIO;
-	}
-
-	io = g_io_channel_unix_new(sk);
-	if (io == NULL) {
-		close(sk);
-		return -ENOMEM;
-	}
-
-	if (g_io_channel_set_flags(io, G_IO_FLAG_NONBLOCK,
-						NULL) != G_IO_STATUS_NORMAL) {
-		g_io_channel_unref(io);
-		close(sk);
-		return -EPERM;
-	}
-
-	memset(&addr, 0, sizeof(addr));
-	addr.rc_family = AF_BLUETOOTH;
-	bacpy(&addr.rc_bdaddr, dst);
-	addr.rc_channel = channel;
-
-	if (connect(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
-		if (errno != EAGAIN && errno != EINPROGRESS) {
-			g_io_channel_unref(io);
-			close(sk);
-			return -EIO;
-		}
-	}
-
-	g_io_add_watch(io, G_IO_OUT | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
-							function, user_data);
-
-	g_io_channel_unref(io);
-
-	return 0;
+	error("%s", err->message);
+	g_error_free(err);
+	return -EIO;
 }
 
 static void search_callback(uint8_t type, uint16_t status,
@@ -520,7 +486,7 @@ void session_shutdown(struct session_data *session)
 {
 	struct transfer_data *transfer;
 
-	debug("session_shutdown(%p)", session);
+	DBG("%p", session);
 	transfer = session->pending ? session->pending->data : NULL;
 
 	session_ref(session);
@@ -800,7 +766,7 @@ static void session_request_reply(DBusPendingCall *call, gpointer user_data)
 {
 	struct session_data *session = user_data;
 	struct agent_data *agent = session->agent;
-	struct agent_pending *pending = agent->pending;
+	struct pending_data *pending = agent->pending;
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	const char *name;
 	DBusError derr;
@@ -810,14 +776,16 @@ static void session_request_reply(DBusPendingCall *call, gpointer user_data)
 		error("Replied with an error: %s, %s",
 				derr.name, derr.message);
 		dbus_error_free(&derr);
-		goto fail;
+		dbus_message_unref(reply);
+		transfer_unregister(pending->transfer);
+		return;
 	}
 
 	dbus_message_get_args(reply, NULL,
 			DBUS_TYPE_STRING, &name,
 			DBUS_TYPE_INVALID);
 
-	debug("Agent.Request() reply: %s", name);
+	DBG("Agent.Request() reply: %s", name);
 
 	if (strlen(name)) {
 		g_free(pending->transfer->name);
@@ -825,14 +793,22 @@ static void session_request_reply(DBusPendingCall *call, gpointer user_data)
 	}
 
 	pending->cb(session, pending->transfer);
-
+	dbus_message_unref(reply);
 	free_pending(pending);
 	agent->pending = NULL;
 
 	return;
+}
 
-fail:
-	transfer_unregister(pending->transfer);
+static gboolean session_request_proceed(gpointer data)
+{
+	struct pending_data *pending = data;
+	struct transfer_data *transfer = pending->transfer;
+
+	pending->cb(transfer->session, transfer);
+	free_pending(pending);
+
+	return FALSE;
 }
 
 static int session_request(struct session_data *session, session_callback_t cb,
@@ -840,11 +816,14 @@ static int session_request(struct session_data *session, session_callback_t cb,
 {
 	struct agent_data *agent = session->agent;
 	DBusMessage *message;
-	DBusPendingCall *call;
-	struct agent_pending *pending;
+	struct pending_data *pending;
+
+	pending = g_new0(struct pending_data, 1);
+	pending->cb = cb;
+	pending->transfer = transfer;
 
 	if (agent == NULL || transfer->path == NULL) {
-		cb(session, transfer);
+		g_idle_add(session_request_proceed, pending);
 		return 0;
 	}
 
@@ -855,23 +834,20 @@ static int session_request(struct session_data *session, session_callback_t cb,
 			DBUS_TYPE_OBJECT_PATH, &transfer->path,
 			DBUS_TYPE_INVALID);
 
-
-	if (!dbus_connection_send_with_reply(session->conn, message, &call, -1)) {
+	if (!dbus_connection_send_with_reply(session->conn, message,
+						&pending->call, -1)) {
 		dbus_message_unref(message);
 		return -ENOMEM;
 	}
 
-	dbus_message_unref(message);
-
-	pending = g_new0(struct agent_pending, 1);
-	pending->call = call;
-	pending->cb = cb;
-	pending->transfer = transfer;
 	agent->pending = pending;
 
-	dbus_pending_call_set_notify(call, session_request_reply, session, NULL);
+	dbus_message_unref(message);
 
-	debug("Agent.Request(\"%s\")", transfer->path);
+	dbus_pending_call_set_notify(pending->call, session_request_reply,
+					session, NULL);
+
+	DBG("Agent.Request(\"%s\")", transfer->path);
 
 	return 0;
 }
@@ -921,7 +897,7 @@ static void session_notify_complete(struct session_data *session,
 
 done:
 
-	debug("Transfer(%p) complete", transfer);
+	DBG("Transfer(%p) complete", transfer);
 
 	session_terminate_transfer(session, transfer);
 }
@@ -1004,7 +980,7 @@ static void session_notify_progress(struct session_data *session,
 	g_dbus_send_message(session->conn, message);
 
 done:
-	debug("Transfer(%p) progress: %ld bytes", transfer,
+	DBG("Transfer(%p) progress: %ld bytes", transfer,
 			(long int ) transferred);
 
 	if (transferred == transfer->size)
@@ -1031,31 +1007,43 @@ fail:
 static void session_prepare_get(struct session_data *session, void *data)
 {
 	struct transfer_data *transfer = data;
+	int err;
 
-	if (transfer_get(transfer, transfer_progress, session) < 0)
-		transfer_unregister(transfer);
+	err = transfer_get(transfer, transfer_progress, session);
+	if (err < 0) {
+		session_notify_error(session, transfer, strerror(-err));
+		return;
+	}
+
+	DBG("Transfer(%p) started", transfer);
 }
 
 int session_get(struct session_data *session, const char *type,
 		const char *filename, const char *targetname,
-		const guint8  *apparam, gint apparam_size,
+		const guint8 *apparam, gint apparam_size,
 		session_callback_t func)
 {
 	struct transfer_data *transfer;
-	struct transfer_params *params;
+	struct transfer_params *params = NULL;
 	int err;
 
 	if (session->obex == NULL)
 		return -ENOTCONN;
 
-	params = g_new0(struct transfer_params, 1);
-	params->data = apparam;
-	params->size = apparam_size;
+	if (apparam != NULL) {
+		params = g_new0(struct transfer_params, 1);
+		params->data = g_new(guint8, apparam_size);
+		memcpy(params->data, apparam, apparam_size);
+		params->size = apparam_size;
+	}
 
 	transfer = transfer_register(session, filename, targetname, type,
 					params);
 	if (transfer == NULL) {
-		g_free(params);
+		if (params != NULL) {
+			g_free(params->data);
+			g_free(params);
+		}
 		return -EIO;
 	}
 
@@ -1278,7 +1266,7 @@ int session_pull(struct session_data *session,
 	if (session->obex == NULL)
 		return -ENOTCONN;
 
-	transfer = transfer_register(session, NULL, NULL, type, NULL);
+	transfer = transfer_register(session, NULL, filename, type, NULL);
 	if (transfer == NULL) {
 		return -EIO;
 	}
@@ -1332,7 +1320,7 @@ int session_register(struct session_data *session)
 		return -EIO;
 	}
 
-	debug("Session(%p) registered %s", session, session->path);
+	DBG("Session(%p) registered %s", session, session->path);
 
 	return 0;
 }
@@ -1350,9 +1338,15 @@ void session_set_data(struct session_data *session, void *priv)
 static void session_prepare_put(struct session_data *session, void *data)
 {
 	struct transfer_data *transfer = data;
+	int err;
 
-	if (transfer_put(transfer, transfer_progress, session) < 0)
-		transfer_unregister(transfer);
+	err = transfer_put(transfer, transfer_progress, session);
+	if (err < 0) {
+		session_notify_error(session, transfer, strerror(-err));
+		return;
+	}
+
+	DBG("Transfer(%p) started", transfer);
 }
 
 int session_put(struct session_data *session, char *buf, const char *targetname)
