@@ -45,7 +45,7 @@
 #include <openobex/obex_const.h>
 
 #include "plugin.h"
-#include "logging.h"
+#include "log.h"
 #include "obex.h"
 #include "mimetype.h"
 #include "service.h"
@@ -58,7 +58,7 @@
 #define FL_TYPE "<!DOCTYPE folder-listing SYSTEM \"obex-folder-listing.dtd\">" EOL_CHARS
 
 #define FL_TYPE_PCSUITE "<!DOCTYPE folder-listing SYSTEM \"obex-folder-listing.dtd\"" EOL_CHARS \
-                        "  [ <!ATTLIST folder mem-type CDATA #IMPLIED> ]>" EOL_CHARS
+			"  [ <!ATTLIST folder mem-type CDATA #IMPLIED> ]>" EOL_CHARS
 
 #define FL_BODY_BEGIN "<folder-listing version=\"1.0\">" EOL_CHARS
 
@@ -77,16 +77,22 @@
 			" modified=\"%s\" mem-type=\"DEV\"" \
 			" created=\"%s\"/>" EOL_CHARS
 
-static const guint8 FTP_TARGET[TARGET_SIZE] = {
+static const uint8_t FTP_TARGET[TARGET_SIZE] = {
 			0xF9, 0xEC, 0x7B, 0xC4,  0x95, 0x3C, 0x11, 0xD2,
 			0x98, 0x4E, 0x52, 0x54,  0x00, 0xDC, 0x9E, 0x09  };
 
-static gchar *file_stat_line(gchar *filename, struct stat *fstat,
-				struct stat *dstat, gboolean root,
-				gboolean pcsuite)
+#define PCSUITE_WHO_SIZE 8
+
+static const uint8_t PCSUITE_WHO[PCSUITE_WHO_SIZE] = {
+			'P','C',' ','S','u','i','t','e' };
+
+
+static char *file_stat_line(char *filename, struct stat *fstat,
+					struct stat *dstat, gboolean root,
+					gboolean pcsuite)
 {
-	gchar perm[51], atime[18], ctime[18], mtime[18];
-	gchar *escaped, *ret = NULL;
+	char perm[51], atime[18], ctime[18], mtime[18];
+	char *escaped, *ret = NULL;
 
 	snprintf(perm, 50, "user-perm=\"%s%s%s\" group-perm=\"%s%s%s\" "
 			"other-perm=\"%s%s%s\"",
@@ -113,22 +119,26 @@ static gchar *file_stat_line(gchar *filename, struct stat *fstat,
 						mtime, ctime);
 		else
 			ret = g_strdup_printf(FL_FOLDER_ELEMENT, escaped, perm,
-						atime, mtime, ctime);
+							atime, mtime, ctime);
 	} else if (S_ISREG(fstat->st_mode))
 		ret = g_strdup_printf(FL_FILE_ELEMENT, escaped, fstat->st_size,
-					perm, atime, mtime, ctime);
+						perm, atime, mtime, ctime);
 
 	g_free(escaped);
 
 	return ret;
 }
 
-static gpointer filesystem_open(const char *name, int oflag, mode_t mode,
-		gpointer context, size_t *size, int *err)
+static void *filesystem_open(const char *name, int oflag, mode_t mode,
+					void *context, size_t *size, int *err)
 {
 	struct stat stats;
 	struct statvfs buf;
+	const char *root_folder;
+	char *folder;
+	gboolean root;
 	int fd = open(name, oflag, mode);
+	uint64_t avail;
 
 	if (fd < 0) {
 		if (err)
@@ -142,6 +152,21 @@ static gpointer filesystem_open(const char *name, int oflag, mode_t mode,
 		goto failed;
 	}
 
+	root_folder = obex_option_root_folder();
+	folder = g_path_get_dirname(name);
+	root = g_strcmp0(folder, root_folder);
+
+	g_free(folder);
+
+	if (!root || obex_option_symlinks()) {
+		if (S_ISLNK(stats.st_mode)) {
+			if (err)
+				*err = -EPERM;
+			goto failed;
+		}
+
+	}
+
 	if (oflag == O_RDONLY) {
 		if (size)
 			*size = stats.st_size;
@@ -157,7 +182,8 @@ static gpointer filesystem_open(const char *name, int oflag, mode_t mode,
 	if (size == NULL)
 		goto done;
 
-	if (buf.f_bsize * buf.f_bavail < *size) {
+	avail = (uint64_t) buf.f_bsize * buf.f_bavail;
+	if (avail < *size) {
 		if (err)
 			*err = -ENOSPC;
 		goto failed;
@@ -174,7 +200,7 @@ failed:
 	return NULL;
 }
 
-static int filesystem_close(gpointer object)
+static int filesystem_close(void *object)
 {
 	if (close(GPOINTER_TO_INT(object)) < 0)
 		return -errno;
@@ -182,7 +208,8 @@ static int filesystem_close(gpointer object)
 	return 0;
 }
 
-static ssize_t filesystem_read(gpointer object, void *buf, size_t count)
+static ssize_t filesystem_read(void *object, void *buf, size_t count,
+								uint8_t *hi)
 {
 	ssize_t ret;
 
@@ -190,10 +217,12 @@ static ssize_t filesystem_read(gpointer object, void *buf, size_t count)
 	if (ret < 0)
 		return -errno;
 
+	*hi = OBEX_HDR_BODY;
+
 	return ret;
 }
 
-static ssize_t filesystem_write(gpointer object, const void *buf, size_t count)
+static ssize_t filesystem_write(void *object, const void *buf, size_t count)
 {
 	ssize_t ret;
 
@@ -208,16 +237,19 @@ struct capability_object {
 	int pid;
 	int output;
 	int err;
-	guint watch;
+	unsigned int watch;
 	GString *buffer;
 };
 
-static void script_exited(GPid pid, gint status, gpointer data)
+static void script_exited(GPid pid, int status, void *data)
 {
 	struct capability_object *object = data;
 	char buf[128];
 
 	object->pid = -1;
+	object->watch = 0;
+
+	DBG("pid: %d status: %d", pid, status);
 
 	if (WEXITSTATUS(status) != EXIT_SUCCESS) {
 		memset(buf, 0, sizeof(buf));
@@ -243,14 +275,16 @@ static int capability_exec(const char **argv, int *output, int *err)
 		return -EPERM;
 	}
 
+	DBG("executing %s pid %d", argv[0], pid);
+
 	return pid;
 }
 
-static gpointer capability_open(const char *name, int oflag, mode_t mode,
-		gpointer context, size_t *size, int *err)
+static void *capability_open(const char *name, int oflag, mode_t mode,
+					void *context, size_t *size, int *err)
 {
 	struct capability_object *object = NULL;
-	gchar *buf;
+	char *buf;
 	const char *argv[2];
 
 	if (oflag != O_RDONLY)
@@ -289,9 +323,6 @@ static gpointer capability_open(const char *name, int oflag, mode_t mode,
 
 	object->watch = g_child_watch_add(object->pid, script_exited, object);
 
-	if (size)
-		*size = 1;
-
 done:
 	if (err)
 		*err = 0;
@@ -306,25 +337,26 @@ fail:
 	return NULL;
 }
 
-static gpointer folder_open(const char *name, int oflag, mode_t mode,
-		 gpointer context, size_t *size, int *err)
+static GString *append_pcsuite_preamble(GString *object)
+{
+	return g_string_append(object, FL_TYPE_PCSUITE);
+}
+
+static GString *append_folder_preamble(GString *object)
+{
+	return g_string_append(object, FL_TYPE);
+}
+
+static GString *append_listing(GString *object, const char *name,
+				gboolean pcsuite, size_t *size, int *err)
 {
-	struct obex_session *os = context;
 	struct stat fstat, dstat;
 	struct dirent *ep;
-	GString *object;
 	DIR *dp;
-	gboolean root, pcsuite, symlinks;
-	gint ret;
+	gboolean root, symlinks;
+	int ret;
 
-	pcsuite = obex_get_service(os) & OBEX_PCSUITE ? TRUE : FALSE;
-
-	object = g_string_new(FL_VERSION);
-	object = g_string_append(object, pcsuite ? FL_TYPE_PCSUITE : FL_TYPE);
-
-	object = g_string_append(object, FL_BODY_BEGIN);
-
-	root = g_str_equal(name, obex_get_root_folder(os));
+	root = g_str_equal(name, obex_option_root_folder());
 
 	dp = opendir(name);
 	if (dp == NULL) {
@@ -333,7 +365,7 @@ static gpointer folder_open(const char *name, int oflag, mode_t mode,
 		goto failed;
 	}
 
-	symlinks = obex_get_symlinks(os);
+	symlinks = obex_option_symlinks();
 	if (root && symlinks)
 		ret = stat(name, &dstat);
 	else {
@@ -348,9 +380,9 @@ static gpointer folder_open(const char *name, int oflag, mode_t mode,
 	}
 
 	while ((ep = readdir(dp))) {
-		gchar *filename;
-		gchar *fullname;
-		gchar *line;
+		char *filename;
+		char *fullname;
+		char *line;
 
 		if (ep->d_name[0] == '.')
 			continue;
@@ -369,7 +401,7 @@ static gpointer folder_open(const char *name, int oflag, mode_t mode,
 			ret = lstat(fullname, &fstat);
 
 		if (ret < 0) {
-			debug("%s: %s(%d)", root ? "stat" : "lstat",
+			DBG("%s: %s(%d)", root ? "stat" : "lstat",
 					strerror(errno), errno);
 			g_free(filename);
 			g_free(fullname);
@@ -378,7 +410,7 @@ static gpointer folder_open(const char *name, int oflag, mode_t mode,
 
 		g_free(fullname);
 
-		line = file_stat_line(filename, &fstat, &dstat, root, pcsuite);
+		line = file_stat_line(filename, &fstat, &dstat, root, FALSE);
 		if (line == NULL) {
 			g_free(filename);
 			continue;
@@ -409,7 +441,31 @@ failed:
 	return NULL;
 }
 
-int string_free(gpointer object)
+static void *folder_open(const char *name, int oflag, mode_t mode,
+					void *context, size_t *size, int *err)
+{
+	GString *object;
+
+	object = g_string_new(FL_VERSION);
+	object = append_folder_preamble(object);
+	object = g_string_append(object, FL_BODY_BEGIN);
+
+	return append_listing(object, name, FALSE, size, err);
+}
+
+static void *pcsuite_open(const char *name, int oflag, mode_t mode,
+					void *context, size_t *size, int *err)
+{
+	GString *object;
+
+	object = g_string_new(FL_VERSION);
+	object = append_pcsuite_preamble(object);
+	object = g_string_append(object, FL_BODY_BEGIN);
+
+	return append_listing(object, name, TRUE, size, err);
+}
+
+int string_free(void *object)
 {
 	GString *string = object;
 
@@ -418,7 +474,7 @@ int string_free(gpointer object)
 	return 0;
 }
 
-ssize_t string_read(gpointer object, void *buf, size_t count)
+ssize_t string_read(void *object, void *buf, size_t count)
 {
 	GString *string = object;
 	ssize_t len;
@@ -433,10 +489,19 @@ ssize_t string_read(gpointer object, void *buf, size_t count)
 	return len;
 }
 
-static ssize_t capability_read(gpointer object, void *buf, size_t count)
+static ssize_t folder_read(void *object, void *buf, size_t count, uint8_t *hi)
+{
+	*hi = OBEX_HDR_BODY;
+	return string_read(object, buf, count);
+}
+
+static ssize_t capability_read(void *object, void *buf, size_t count,
+								uint8_t *hi)
 {
 	struct capability_object *obj = object;
 
+	*hi = OBEX_HDR_BODY;
+
 	if (obj->buffer)
 		return string_read(obj->buffer, buf, count);
 
@@ -446,25 +511,36 @@ static ssize_t capability_read(gpointer object, void *buf, size_t count)
 	return read(obj->output, buf, count);
 }
 
-static int capability_close(gpointer object)
+static int capability_close(void *object)
 {
 	struct capability_object *obj = object;
+	int err = 0;
 
-	if (obj->pid >= 0) {
+	if (obj->pid < 0)
+		goto done;
+
+	if (obj->watch)
 		g_source_remove(obj->watch);
-		kill(obj->pid, SIGTERM);
-		g_spawn_close_pid(obj->pid);
+
+	g_spawn_close_pid(obj->pid);
+
+	DBG("kill: pid %d", obj->pid);
+	err = kill(obj->pid, SIGTERM);
+	if (err < 0) {
+		err = -errno;
+		error("kill: %s (%d)", strerror(-err), -err);
 	}
 
+done:
 	if (obj->buffer != NULL)
 		g_string_free(obj->buffer, TRUE);
 
 	g_free(obj);
 
-	return 0;
+	return err;
 }
 
-struct obex_mime_type_driver file = {
+static struct obex_mime_type_driver file = {
 	.open = filesystem_open,
 	.close = filesystem_close,
 	.read = filesystem_read,
@@ -472,7 +548,7 @@ struct obex_mime_type_driver file = {
 	.remove = remove,
 };
 
-struct obex_mime_type_driver capability = {
+static struct obex_mime_type_driver capability = {
 	.target = FTP_TARGET,
 	.mimetype = "x-obex/capability",
 	.open = capability_open,
@@ -480,12 +556,22 @@ struct obex_mime_type_driver capability = {
 	.read = capability_read,
 };
 
-struct obex_mime_type_driver folder = {
+static struct obex_mime_type_driver folder = {
 	.target = FTP_TARGET,
 	.mimetype = "x-obex/folder-listing",
 	.open = folder_open,
 	.close = string_free,
-	.read = string_read,
+	.read = folder_read,
+};
+
+static struct obex_mime_type_driver pcsuite = {
+	.target = FTP_TARGET,
+	.who = PCSUITE_WHO,
+	.who_size = PCSUITE_WHO_SIZE,
+	.mimetype = "x-obex/folder-listing",
+	.open = pcsuite_open,
+	.close = string_free,
+	.read = folder_read,
 };
 
 static int filesystem_init(void)
@@ -500,6 +586,10 @@ static int filesystem_init(void)
 	if (err < 0)
 		return err;
 
+	err = obex_mime_type_driver_register(&pcsuite);
+	if (err < 0)
+		return err;
+
 	return obex_mime_type_driver_register(&file);
 }
 
