@@ -22,13 +22,15 @@
 
 #include <string.h>
 #include <stdlib.h>
+#include <time.h>
+#include <stdio.h>
 #include <errno.h>
 #include <glib.h>
 #include <dbus/dbus.h>
 #include <openobex/obex.h>
 #include <openobex/obex_const.h>
 
-#include "logging.h"
+#include "log.h"
 #include "obex.h"
 #include "service.h"
 #include "mimetype.h"
@@ -49,26 +51,27 @@
 	"nco:nameHonorificSuffix(?c) nco:emailAddress(?e) "		\
 	"nco:phoneNumber(?w) nco:pobox(?p) nco:extendedAddress(?p) "	\
 	"nco:streetAddress(?p) nco:locality(?p) nco:region(?p) "	\
-	"nco:postalcode(?p) nco:country(?p) "				\
+	"nco:postalcode(?p) nco:country(?p) \"NOTACALL\" \"false\" "	\
+	"\"false\" ?c "							\
 	"WHERE { "							\
-		"?c a nco:PersonContact ; "				\
-		"nco:hasPhoneNumber ?h . "				\
+		"?c a nco:PersonContact . "				\
+	"OPTIONAL { ?c nco:hasPhoneNumber ?h . } "			\
+	"OPTIONAL { ?c nco:hasEmailAddress ?e . } "			\
+	"OPTIONAL { ?c nco:hasPostalAddress ?p . } "			\
 	"OPTIONAL { "							\
 		"?c nco:hasAffiliation ?a . "				\
 		"?a nco:hasPhoneNumber ?w . "				\
-		"?c nco:hasEmailAddress ?e . "				\
-		"?c nco:hasPostalAddress ?p . "				\
 	"} "								\
 	"}"
 
 #define CONTACTS_QUERY_ALL_LIST						\
-	"SELECT ?c nco:contactUID(?c) nco:nameFamily(?c) "		\
+	"SELECT ?c nco:nameFamily(?c) "					\
 	"nco:nameGiven(?c) nco:nameAdditional(?c) "			\
 	"nco:nameHonorificPrefix(?c) nco:nameHonorificSuffix(?c) "	\
 	"nco:phoneNumber(?h) "						\
 	"WHERE { "							\
-		"?c a nco:PersonContact ; "				\
-		"nco:hasPhoneNumber ?h . "				\
+		"?c a nco:PersonContact . "				\
+	"OPTIONAL { ?c nco:hasPhoneNumber ?h . } "			\
 	"}"
 
 #define MISSED_CALLS_QUERY						\
@@ -78,24 +81,25 @@
 	"nco:nameHonorificSuffix(?c) nco:emailAddress(?e) "		\
 	"nco:phoneNumber(?w) nco:pobox(?p) nco:extendedAddress(?p) "	\
 	"nco:streetAddress(?p) nco:locality(?p) nco:region(?p) "	\
-	"nco:postalcode(?p) nco:country(?p) "				\
+	"nco:postalcode(?p) nco:country(?p) nmo:receivedDate(?call) "	\
+	"nmo:isSent(?call) nmo:isAnswered(?call) ?c "			\
 	"WHERE { "							\
 		"?call a nmo:Call ; "					\
 		"nmo:from ?c ; "					\
 		"nmo:isSent false ; "					\
 		"nmo:isAnswered false ."				\
-		"?c a nco:PersonContact ; "				\
-		"nco:hasPhoneNumber ?h . "				\
+		"?c a nco:Contact . "					\
+	"OPTIONAL { ?c nco:hasPhoneNumber ?h . } "			\
+	"OPTIONAL { ?c nco:hasEmailAddress ?e . } "			\
+	"OPTIONAL { ?c nco:hasPostalAddress ?p . } "			\
 	"OPTIONAL { "							\
 		"?c nco:hasAffiliation ?a . "				\
 		"?a nco:hasPhoneNumber ?w . "				\
-		"?c nco:hasEmailAddress ?e . "				\
-		"?c nco:hasPostalAddress ?p . "				\
 	"} "								\
 	"} ORDER BY DESC(nmo:receivedDate(?call))"
 
 #define MISSED_CALLS_LIST						\
-	"SELECT ?c nco:contactUID(?c) nco:nameFamily(?c) "		\
+	"SELECT ?c nco:nameFamily(?c) "					\
 	"nco:nameGiven(?c) nco:nameAdditional(?c) "			\
 	"nco:nameHonorificPrefix(?c) nco:nameHonorificSuffix(?c) "	\
 	"nco:phoneNumber(?h) "						\
@@ -104,8 +108,8 @@
 		"nmo:from ?c ; "					\
 		"nmo:isSent false ; "					\
 		"nmo:isAnswered false ."				\
-		"?c a nco:PersonContact ; "				\
-		"nco:hasPhoneNumber ?h . "				\
+		"?c a nco:Contact . "					\
+	"OPTIONAL { ?c nco:hasPhoneNumber ?h . } "			\
 	"} ORDER BY DESC(nmo:receivedDate(?call))"
 
 #define INCOMING_CALLS_QUERY						\
@@ -115,23 +119,24 @@
 	"nco:nameHonorificSuffix(?c) nco:emailAddress(?e) "		\
 	"nco:phoneNumber(?w) nco:pobox(?p) nco:extendedAddress(?p) "	\
 	"nco:streetAddress(?p) nco:locality(?p) nco:region(?p) "	\
-	"nco:postalcode(?p) nco:country(?p) "				\
+	"nco:postalcode(?p) nco:country(?p) nmo:receivedDate(?call) "	\
+	"nmo:isSent(?call) nmo:isAnswered(?call) ?c "			\
 	"WHERE { "							\
 		"?call a nmo:Call ; "					\
 		"nmo:from ?c ; "					\
 		"nmo:isSent false . "					\
-		"?c a nco:PersonContact ; "				\
-		"nco:hasPhoneNumber ?h . "				\
+		"?c a nco:Contact . "					\
+	"OPTIONAL { ?c nco:hasPhoneNumber ?h . } "			\
+	"OPTIONAL { ?c nco:hasEmailAddress ?e . } "			\
+	"OPTIONAL { ?c nco:hasPostalAddress ?p . } "			\
 	"OPTIONAL { "							\
 		"?c nco:hasAffiliation ?a . "				\
 		"?a nco:hasPhoneNumber ?w . "				\
-		"?c nco:hasEmailAddress ?e . "				\
-		"?c nco:hasPostalAddress ?p . "				\
 	"} "								\
 	"} ORDER BY DESC(nmo:receivedDate(?call))"
 
 #define INCOMING_CALLS_LIST						\
-	"SELECT ?c nco:contactUID(?c) nco:nameFamily(?c) "		\
+	"SELECT ?c nco:nameFamily(?c) "					\
 	"nco:nameGiven(?c) nco:nameAdditional(?c) "			\
 	"nco:nameHonorificPrefix(?c) nco:nameHonorificSuffix(?c) "	\
 	"nco:phoneNumber(?h) "						\
@@ -139,8 +144,8 @@
 		"?call a nmo:Call ; "					\
 		"nmo:from ?c ; "					\
 		"nmo:isSent false . "					\
-		"?c a nco:PersonContact ; "				\
-		"nco:hasPhoneNumber ?h . "				\
+		"?c a nco:Contact . "					\
+	"OPTIONAL { ?c nco:hasPhoneNumber ?h . } "			\
 	"} ORDER BY DESC(nmo:receivedDate(?call))"
 
 #define OUTGOING_CALLS_QUERY						\
@@ -150,23 +155,24 @@
 	"nco:nameHonorificSuffix(?c) nco:emailAddress(?e) "		\
 	"nco:phoneNumber(?w) nco:pobox(?p) nco:extendedAddress(?p) "	\
 	"nco:streetAddress(?p) nco:locality(?p) nco:region(?p) "	\
-	"nco:postalcode(?p) nco:country(?p) "				\
+	"nco:postalcode(?p) nco:country(?p) nmo:receivedDate(?call) "	\
+	"nmo:isSent(?call) nmo:isAnswered(?call) ?c "			\
 	"WHERE { "							\
 		"?call a nmo:Call ; "					\
 		"nmo:to ?c ; "						\
 		"nmo:isSent true . "					\
-		"?c a nco:PersonContact ; "				\
-		"nco:hasPhoneNumber ?h . "				\
+		"?c a nco:Contact . "					\
+	"OPTIONAL { ?c nco:hasPhoneNumber ?h . } "			\
+	"OPTIONAL { ?c nco:hasEmailAddress ?e . } "			\
+	"OPTIONAL { ?c nco:hasPostalAddress ?p . } "			\
 	"OPTIONAL { "							\
 		"?c nco:hasAffiliation ?a . "				\
 		"?a nco:hasPhoneNumber ?w . "				\
-		"?c nco:hasEmailAddress ?e . "				\
-		"?c nco:hasPostalAddress ?p . "				\
 	"} "								\
 	"} ORDER BY DESC(nmo:sentDate(?call))"
 
 #define OUTGOING_CALLS_LIST						\
-	"SELECT ?c nco:contactUID(?c) nco:nameFamily(?c) "		\
+	"SELECT ?c nco:nameFamily(?c) "					\
 	"nco:nameGiven(?c) nco:nameAdditional(?c) "			\
 	"nco:nameHonorificPrefix(?c) nco:nameHonorificSuffix(?c) "	\
 	"nco:phoneNumber(?h) "						\
@@ -174,8 +180,8 @@
 		"?call a nmo:Call ; "					\
 		"nmo:to ?c ; "						\
 		"nmo:isSent true . "					\
-		"?c a nco:PersonContact ; "				\
-		"nco:hasPhoneNumber ?h . "				\
+		"?c a nco:Contact . "					\
+	"OPTIONAL { ?c nco:hasPhoneNumber ?h . } "				\
 	"} ORDER BY DESC(nmo:sentDate(?call))"
 
 #define COMBINED_CALLS_QUERY						\
@@ -185,72 +191,72 @@
 	"nco:nameHonorificSuffix(?c) nco:emailAddress(?e) "		\
 	"nco:phoneNumber(?w) nco:pobox(?p) nco:extendedAddress(?p) "	\
 	"nco:streetAddress(?p) nco:locality(?p) nco:region(?p) "	\
-	"nco:postalcode(?p) nco:country(?p) "				\
+	"nco:postalcode(?p) nco:country(?p) nmo:receivedDate(?call) "	\
+	"nmo:isSent(?call) nmo:isAnswered(?call) ?c "			\
 	"WHERE { "							\
 	"{ "								\
 		"?call a nmo:Call ; "					\
 		"nmo:to ?c ; "						\
 		"nmo:isSent true . "					\
-		"?c a nco:PersonContact ; "				\
-		"nco:hasPhoneNumber ?h . "				\
+		"?c a nco:Contact . "					\
+		"OPTIONAL { ?c nco:hasPhoneNumber ?h . } "		\
+		"OPTIONAL { ?c nco:hasEmailAddress ?e . } "		\
+		"OPTIONAL { ?c nco:hasPostalAddress ?p . } "		\
 		"OPTIONAL { "						\
 			"?c nco:hasAffiliation ?a . "			\
 			"?a nco:hasPhoneNumber ?w . "			\
-			"?c nco:hasEmailAddress ?e . "			\
-			"?c nco:hasPostalAddress ?p . "			\
 		"} "							\
 	"} UNION { "							\
 		"?call a nmo:Call ; "					\
 		"nmo:from ?c ; "					\
 		"nmo:isSent false . "					\
-		"?c a nco:PersonContact ; "				\
-		"nco:hasPhoneNumber ?h . "				\
+		"?c a nco:Contact . "					\
+		"OPTIONAL { ?c nco:hasPhoneNumber ?h . } "		\
+		"OPTIONAL { ?c nco:hasEmailAddress ?e . } "		\
+		"OPTIONAL { ?c nco:hasPostalAddress ?p . } "		\
 		"OPTIONAL { "						\
 			"?c nco:hasAffiliation ?a . "			\
 			"?a nco:hasPhoneNumber ?w . "			\
-			"?c nco:hasEmailAddress ?e . "			\
-			"?c nco:hasPostalAddress ?p . "			\
 		"} "							\
-	"} } "
+	"} } ORDER BY DESC(nmo:receivedDate(?call))"
 
 #define COMBINED_CALLS_LIST						\
-	"SELECT ?c nco:contactUID(?c) nco:nameFamily(?c) "		\
-	"nco:nameGiven(?c) nco:nameAdditional(?c) "			\
-	"nco:nameHonorificPrefix(?c) nco:nameHonorificSuffix(?c) "	\
-	"nco:phoneNumber(?h) "						\
+	"SELECT ?c nco:nameFamily(?c) nco:nameGiven(?c) "		\
+	"nco:nameAdditional(?c) nco:nameHonorificPrefix(?c) "		\
+	"nco:nameHonorificSuffix(?c) nco:phoneNumber(?h) "		\
 	"WHERE { "							\
 	"{ "								\
 		"?call a nmo:Call ; "					\
 		"nmo:to ?c ; "						\
 		"nmo:isSent true . "					\
-		"?c a nco:PersonContact ; "				\
-		"nco:hasPhoneNumber ?h . "				\
+		"?c a nco:Contact . "					\
+		"OPTIONAL { ?c nco:hasPhoneNumber ?h . } "		\
 	"} UNION { "							\
 		"?call a nmo:Call ; "					\
 		"nmo:from ?c ; "					\
 		"nmo:isSent false . "					\
-		"?c a nco:PersonContact ; "				\
-		"nco:hasPhoneNumber ?h . "				\
-	"} } "
+		"?c a nco:Contact . "					\
+		"OPTIONAL { ?c nco:hasPhoneNumber ?h . } "		\
+	"} } ORDER BY DESC(nmo:receivedDate(?call))"
 
 
 #define CONTACTS_QUERY_FROM_URI						\
-	"SELECT nco:phoneNumber(?h) nco:fullname(?c) "			\
-	"nco:nameFamily(?c) nco:nameGiven(?c) nco:nameAdditional(?c) "	\
-	"nco:nameHonorificPrefix(?c) nco:nameHonorificSuffix(?c)  "	\
-	"nco:emailAddress(?e) "						\
+	"SELECT nco:phoneNumber(?h) nco:fullname(<%s>) "		\
+	"nco:nameFamily(<%s>) nco:nameGiven(<%s>) "			\
+	"nco:nameAdditional(<%s>) nco:nameHonorificPrefix(<%s>) "	\
+	"nco:nameHonorificSuffix(<%s>) nco:emailAddress(?e) "		\
 	"nco:phoneNumber(?w) nco:pobox(?p) nco:extendedAddress(?p) "	\
 	"nco:streetAddress(?p) nco:locality(?p) nco:region(?p) "	\
-	"nco:postalcode(?p) nco:country(?p) "				\
+	"nco:postalcode(?p) nco:country(?p) \"NOTACALL\" \"false\" "	\
+	"\"false\" <%s> "							\
 	"WHERE { "							\
-		"?c a nco:PersonContact ; "				\
-		"nco:contactUID <%s> ; "				\
-		"nco:hasPhoneNumber ?h . "				\
+		"<%s> a nco:Contact . "					\
+	"OPTIONAL { <%s> nco:hasPhoneNumber ?h . } "			\
+	"OPTIONAL { <%s> nco:hasEmailAddress ?e . } "			\
+	"OPTIONAL { <%s> nco:hasPostalAddress ?p . } "			\
 	"OPTIONAL { "							\
-		"?c nco:hasAffiliation ?a . "				\
+		"<%s> nco:hasAffiliation ?a . "				\
 		"?a nco:hasPhoneNumber ?w . "				\
-		"?c nco:hasEmailAddress ?e . "				\
-		"?c nco:hasPostalAddress ?p . "				\
 	"} "								\
 	"}"
 
@@ -421,6 +427,7 @@ static void query_reply(DBusPendingCall *call, void *user_data)
 		node = string_array_from_iter(element, pending->num_fields);
 		pending->callback(node, pending->num_fields,
 							pending->user_data);
+
 		g_free(node);
 
 		dbus_message_iter_next(&element);
@@ -470,6 +477,68 @@ static int query_tracker(const char *query, int num_fields,
 	return 0;
 }
 
+static char *iso8601_utc_to_localtime(const char *datetime)
+{
+	time_t time;
+	struct tm tm, *local;
+	char localdate[32];
+	char tz;
+	int nr;
+
+	memset(&tm, 0, sizeof(tm));
+
+	nr = sscanf(datetime, "%04u-%02u-%02uT%02u:%02u:%02u%c",
+			&tm.tm_year, &tm.tm_mon, &tm.tm_mday,
+			&tm.tm_hour, &tm.tm_min, &tm.tm_sec,
+			&tz);
+	if (nr < 6) {
+		/* Invalid time format */
+		return g_strdup("");
+	}
+
+	/* Time already in localtime */
+	if (nr == 6)
+		return g_strdup(datetime);
+
+	tm.tm_year -= 1900;	/* Year since 1900 */
+	tm.tm_mon--;		/* Months since January, values 0-11 */
+
+	time = mktime(&tm);
+	time -= timezone;
+
+	local = localtime(&time);
+
+	strftime(localdate, sizeof(localdate), "%Y-%m-%dT%H:%M:%S", local);
+
+	return g_strdup(localdate);
+}
+
+static void set_call_type(struct phonebook_contact *contact,
+				const char *datetime, const char *is_sent,
+				const char *is_answered)
+{
+	gboolean sent, answered;
+
+	if (g_strcmp0(datetime, "NOTACALL") == 0) {
+		contact->calltype = CALL_TYPE_NOT_A_CALL;
+		return;
+	}
+
+	sent = g_str_equal(is_sent, "true");
+	answered = g_str_equal(is_answered, "true");
+
+	if (sent == FALSE) {
+		if (answered == FALSE)
+			contact->calltype = CALL_TYPE_MISSED;
+		else
+			contact->calltype = CALL_TYPE_INCOMING;
+	} else
+		contact->calltype = CALL_TYPE_OUTGOING;
+
+	/* Tracker gives time in the ISO 8601 format, UTC time */
+	contact->datetime = iso8601_utc_to_localtime(datetime);
+}
+
 static void pull_contacts(char **reply, int num_fields, void *user_data)
 {
 	struct phonebook_data *data = user_data;
@@ -477,7 +546,7 @@ static void pull_contacts(char **reply, int num_fields, void *user_data)
 	struct phonebook_contact *contact;
 	struct phonebook_number *number;
 	GString *vcards = data->vcards;
-	int last_index;
+	int last_index, i;
 
 	DBG("reply %p", reply);
 
@@ -488,6 +557,16 @@ static void pull_contacts(char **reply, int num_fields, void *user_data)
 	if (data->vcardentry)
 		goto add_entry;
 
+	/* Last four fields are always present, ignoring them */
+	for (i = 0; i < num_fields - 4; i++) {
+		if (reply[i][0] != '\0')
+			break;
+	}
+
+	if (i == num_fields - 4 &&
+			!g_str_equal(reply[19], TRACKER_DEFAULT_CONTACT_ME))
+		return;
+
 	data->index++;
 
 	/* Just interested in knowing the phonebook size */
@@ -496,7 +575,7 @@ static void pull_contacts(char **reply, int num_fields, void *user_data)
 
 	last_index = params->liststartoffset + params->maxlistcount;
 
-	if (data->index < params->liststartoffset || data->index > last_index)
+	if (data->index <= params->liststartoffset || data->index > last_index)
 		return;
 
 add_entry:
@@ -516,6 +595,8 @@ add_entry:
 	contact->postal = g_strdup(reply[14]);
 	contact->country = g_strdup(reply[15]);
 
+	set_call_type(contact, reply[16], reply[17], reply[18]);
+
 	number = g_new0(struct phonebook_number, 1);
 	number->tel = g_strdup(reply[0]);
 	number->type = 0; /* HOME */
@@ -551,20 +632,35 @@ static void add_to_cache(char **reply, int num_fields, void *user_data)
 {
 	struct cache_data *cache = user_data;
 	char *formatted;
+	int i;
 
 	if (reply == NULL)
 		goto done;
 
-	formatted = g_strdup_printf("%s;%s;%s;%s;%s", reply[2], reply[3],
-						reply[4], reply[5], reply[6]);
+	/* the first element is the URI, always not empty */
+	for (i = 1; i < num_fields; i++) {
+		if (reply[i][0] != '\0')
+			break;
+	}
+
+	if (i == num_fields &&
+			!g_str_equal(reply[0], TRACKER_DEFAULT_CONTACT_ME))
+		return;
+
+	if (i == 6)
+		formatted = g_strdup(reply[6]);
+	else
+		formatted = g_strdup_printf("%s;%s;%s;%s;%s",
+					reply[1], reply[2], reply[3], reply[4],
+					reply[5]);
 
 	/* The owner vCard must have the 0 handle */
 	if (strcmp(reply[0], TRACKER_DEFAULT_CONTACT_ME) == 0)
-		cache->entry_cb(reply[1], 0, formatted, "",
-						reply[7], cache->user_data);
+		cache->entry_cb(reply[0], 0, formatted, "",
+						reply[6], cache->user_data);
 	else
-		cache->entry_cb(reply[1], PHONEBOOK_INVALID_HANDLE, formatted,
-					"", reply[7], cache->user_data);
+		cache->entry_cb(reply[0], PHONEBOOK_INVALID_HANDLE, formatted,
+					"", reply[6], cache->user_data);
 
 	g_free(formatted);
 
@@ -591,7 +687,8 @@ char *phonebook_set_folder(const char *current_folder, const char *new_folder,
 {
 	char *tmp1, *tmp2, *base, *path = NULL;
 	gboolean root, child;
-	int ret, len;
+	int ret = 0;
+	int len;
 
 	root = (g_strcmp0("/", current_folder) == 0);
 	child = (new_folder && strlen(new_folder) != 0);
@@ -647,13 +744,11 @@ char *phonebook_set_folder(const char *current_folder, const char *new_folder,
 	}
 
 done:
-
-	if (!folder_is_valid(path)) {
+	if (ret || !folder_is_valid(path)) {
 		g_free(path);
 		path = NULL;
-		ret = -ENOENT;
-	} else
-		ret = 0;
+		ret = ret ? ret : -ENOENT;
+	}
 
 	if (err)
 		*err = ret;
@@ -679,7 +774,7 @@ int phonebook_pull(const char *name, const struct apparam_field *params,
 	data->user_data = user_data;
 	data->cb = cb;
 
-	return query_tracker(query, 16, pull_contacts, data);
+	return query_tracker(query, 20, pull_contacts, data);
 }
 
 int phonebook_get_entry(const char *folder, const char *id,
@@ -699,9 +794,10 @@ int phonebook_get_entry(const char *folder, const char *id,
 	data->cb = cb;
 	data->vcardentry = TRUE;
 
-	query = g_strdup_printf(CONTACTS_QUERY_FROM_URI, id);
+	query = g_strdup_printf(CONTACTS_QUERY_FROM_URI, id, id, id, id, id,
+						id, id, id, id, id, id, id);
 
-	ret = query_tracker(query, 16, pull_contacts, data);
+	ret = query_tracker(query, 20, pull_contacts, data);
 
 	g_free(query);
 
@@ -725,5 +821,5 @@ int phonebook_create_cache(const char *name, phonebook_entry_cb entry_cb,
 	cache->ready_cb = ready_cb;
 	cache->user_data = user_data;
 
-	return query_tracker(query, 8, add_to_cache, cache);
+	return query_tracker(query, 7, add_to_cache, cache);
 }
